/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include <sys/select.h>
#include <time.h>
#include <stdlib.h>
#include "spamdyke.h"
#include "environment.h"
#include "exec.h"
#include "command_line.h"

int exec_path(char *filename, char *argv[], char *envp[])
  {
  char *path;
  char *tmp_start;
  char *tmp_end;
  char new_filename[MAX_BUF + 1];

  execve(filename, argv, envp);

  if (strchr(filename, '/') == NULL)
    {
    if (((path = find_environment_variable(envp, ENVIRONMENT_PATH, STRLEN(ENVIRONMENT_PATH))) == NULL) ||
        (path[0] == '\0'))
      path = DEFAULT_PATH;

    tmp_start = path;
    tmp_end = NULL;
    while (tmp_start != NULL)
      {
      if ((tmp_end = strchr(tmp_start, ':')) != NULL)
        {
        snprintf(new_filename, MAX_BUF, "%.*s/%s", (int)(tmp_end - tmp_start), tmp_start, filename);
        tmp_start = tmp_end + 1;
        }
      else
        {
        snprintf(new_filename, MAX_BUF, "%s/%s", tmp_start, filename);
        tmp_start = NULL;
        }

      execve(new_filename, argv, envp);
      }
    }

  return(0);
  }

int exec_checkpassword(char *command_line, char *username, char *password, char *timestamp, char *envp[])
  {
  int return_value;
  int i;
  char **child_argv;
  int stdout_pipe[2];
  int stderr_pipe[2];
  int output_pipe[2];
  pid_t child_pid;
  pid_t last_wait;
  char output_buf[MAX_CHECKPASSWORD];
  int strlen_output_buf;
  int strlen_sent_buf;
  int status;
  char *last_arg;
  char *tmp_arg;
  fd_set read_fds;
  fd_set write_fds;
  int max_fd;
  struct timeval timeout;
  time_t start_time;
  char tmp_char;

  return_value = 0;
  child_argv = NULL;
  last_wait = 0;

  if (pipe(output_pipe) != -1)
    if (pipe(stdout_pipe) != -1)
      if (pipe(stderr_pipe) != -1)
        if ((child_pid = fork()) > 0)
          {
          close(output_pipe[0]);
          close(stdout_pipe[1]);
          close(stderr_pipe[1]);

          strlen_output_buf = snprintf(output_buf, MAX_CHECKPASSWORD, "%s%c%s%c%s%c", (username != NULL) ? username : "", '\0', (password != NULL) ? password : "", '\0', (timestamp != NULL) ? timestamp : "", '\0');
          strlen_sent_buf = 0;

          start_time = time(NULL);

          timeout.tv_sec = TIMEOUT_CHECKPASSWORD_SECS - (time(NULL) - start_time);
          timeout.tv_usec = 0;

          FD_ZERO(&read_fds);
          FD_ZERO(&write_fds);
          FD_SET(output_pipe[1], &write_fds);
          FD_SET(stdout_pipe[0], &read_fds);
          FD_SET(stderr_pipe[0], &read_fds);
          max_fd = MAXVAL(MAXVAL(output_pipe[1], stdout_pipe[0]), stderr_pipe[0]);

          while (select(max_fd + 1, &read_fds, &write_fds, NULL, &timeout) >= 0)
            {
            if (FD_ISSET(stdout_pipe[0], &read_fds) &&
                (read(stdout_pipe[0], &tmp_char, 1) == 0))
              break;

            if (FD_ISSET(stderr_pipe[0], &read_fds) &&
                (read(stderr_pipe[0], &tmp_char, 1) == 0))
              break;

            if ((output_pipe[1] >= 0) &&
                FD_ISSET(output_pipe[1], &write_fds))
              {
              write(output_pipe[1], output_buf + strlen_sent_buf, 1);
              strlen_sent_buf++;
              }

            timeout.tv_sec = TIMEOUT_CHECKPASSWORD_SECS - (time(NULL) - start_time);
            timeout.tv_usec = 0;

            if (((last_wait = waitpid(child_pid, &status, WNOHANG)) == 0) &&
                (timeout.tv_sec > 0))
              {
              FD_ZERO(&read_fds);
              FD_ZERO(&write_fds);
              FD_SET(stdout_pipe[0], &read_fds);
              FD_SET(stderr_pipe[0], &read_fds);

              if (strlen_sent_buf < strlen_output_buf)
                {
                FD_SET(output_pipe[1], &write_fds);
                max_fd = MAXVAL(MAXVAL(output_pipe[1], stdout_pipe[0]), stderr_pipe[0]);
                }
              else
                {
                if (output_pipe[1] != -1)
                  {
                  close(output_pipe[1]);
                  output_pipe[1] = -1;
                  }

                timeout.tv_sec = 0;
                timeout.tv_usec = 500;
                max_fd = MAXVAL(stdout_pipe[0], stderr_pipe[0]);
                }
              }
            else
              break;
            }

          if (last_wait == 0)
            waitpid(child_pid, &status, 0);

          if (WIFEXITED(status))
            {
            if (WEXITSTATUS(status) == 0)
              return_value = 1;
            }
          else
            kill(child_pid, SIGKILL);

          if (output_pipe[1] != -1)
            close(output_pipe[1]);
          close(stdout_pipe[0]);
          close(stderr_pipe[0]);
          }
        else if (child_pid == 0)
          {
          close(output_pipe[1]);
          close(stdout_pipe[0]);
          close(stderr_pipe[0]);

          if (dup2(output_pipe[0], CHECKPASSWORD_FD) != -1)
            if (dup2(stdout_pipe[1], STDOUT_FD) != -1)
              if (dup2(stderr_pipe[1], STDERR_FD) != -1)
                {
                last_arg = command_line;
                while ((tmp_arg = strchr(last_arg, ' ')) != NULL)
                  {
                  append_string(&child_argv, last_arg, tmp_arg - last_arg);
                  last_arg = tmp_arg + 1;
                  }
                append_string(&child_argv, last_arg, strlen(last_arg));

                if (!exec_path(child_argv[0], child_argv, envp))
                  {
                  syslog(LOG_ERR, SYSLOG_ERROR_EXEC "%s: %s", child_argv[0], strerror(errno));

                  close(output_pipe[0]);
                  close(stdout_pipe[1]);
                  close(stderr_pipe[1]);

                  for (i = 0; child_argv[i] != NULL; i++)
                    free(child_argv[i]);
                  free(child_argv);

                  exit(111);
                  }
                }
              else
                {
                syslog(LOG_CRIT, SYSLOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

                close(output_pipe[0]);
                close(stdout_pipe[1]);
                close(stderr_pipe[1]);
                }
            else
              {
              syslog(LOG_CRIT, SYSLOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

              close(output_pipe[0]);
              close(stdout_pipe[1]);
              close(stderr_pipe[1]);
              }
          else
            {
            syslog(LOG_CRIT, SYSLOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

            close(output_pipe[0]);
            close(stdout_pipe[1]);
            close(stderr_pipe[1]);
            }
          }
        else
          {
          syslog(LOG_CRIT, SYSLOG_ERROR_FORK "%s", strerror(errno));

          close(stderr_pipe[0]);
          close(stderr_pipe[1]);
          close(stdout_pipe[0]);
          close(stdout_pipe[1]);
          close(output_pipe[0]);
          close(output_pipe[1]);
          }
      else
        {
        syslog(LOG_CRIT, SYSLOG_ERROR_PIPE "%s", strerror(errno));

        close(stdout_pipe[0]);
        close(stdout_pipe[1]);
        close(output_pipe[0]);
        close(output_pipe[1]);
        }
    else
      {
      syslog(LOG_CRIT, SYSLOG_ERROR_PIPE "%s", strerror(errno));

      close(output_pipe[0]);
      close(output_pipe[1]);
      }
  else
    syslog(LOG_CRIT, SYSLOG_ERROR_PIPE "%s", strerror(errno));

  return(return_value);
  }
